from typing import Optional
from aws_alb.alb_traffic_enum import AlbTrafficEnum
from aws_cdk.aws_ec2 import SecurityGroup, IVpc, IPeer, Peer
from aws_cdk.core import Stack
from aws_alb.loadbalancer_sg_modifier import SecurityGroupModifier


class LoadBalancerSecurityGroup(SecurityGroup):
    """
    A class which manages a default security group for a loadbalancer.
    """
    def __init__(
            self,
            scope: Stack,
            prefix: str,
            vpc: IVpc,
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation stack in which the resources should be added.
        :param prefix: A prefix for newly created resources.
        :param vpc: Virtual private cloud in which the security groups and a loadbalancer itself should be placed.
        """
        super().__init__(
            scope=scope,
            id=prefix + 'LBSecurityGroup',
            vpc=vpc,
            allow_all_outbound=False,
            description=f'A {prefix} load balancer security group.',
            security_group_name=prefix + 'AppLoadBalancerSG'
        )

        self.__vpc = vpc
        self.__security_group_modifier = SecurityGroupModifier(self)

    def open_port(self, port: int, peer: Optional[IPeer] = None, ingress: bool = True) -> None:
        """
        Modifies a current security group by opening a specified port.

        :param port: Port to open (allow traffic).
        :param peer: Peer (a CIDR or another security group).
        :param ingress: Specifies whether it is configured for ingress or egress traffic.

        :return: No return.
        """
        if peer:
            self.__security_group_modifier.open_port(port, peer, ingress)

    def get_peer(self, traffic: AlbTrafficEnum) -> Optional[IPeer]:
        """
        Depending on enum creates a peer.

        :param traffic: Configuration enum.

        :return: Peer.
        """
        if traffic == AlbTrafficEnum.INTERNET:
            cidr_peer = Peer.any_ipv4()
        elif traffic == AlbTrafficEnum.VPC:
            cidr_peer = Peer.ipv4(self.__vpc.vpc_cidr_block)
        else:
            cidr_peer = None

        return cidr_peer



from aws_cdk.aws_elasticloadbalancingv2 import CfnTargetGroup
from aws_alb.params.target_group_params import TargetGroupParams
from aws_cdk.core import Stack


class TargetGroupFactory:
    """
    Factory class to create target groups for loadbalancer.
    """
    def __init__(self, scope: Stack) -> None:
        """
        Constructor.

        :param scope: CloudFormation stack instance.
        """
        self.__scope = scope

    def create_target_group(self, target_group_params: TargetGroupParams) -> CfnTargetGroup:
        """
        Creates target group.

        :param target_group_params: Configuration parameters on how to create the target group.

        :return: Target group instance.
        """
        # By default a healthy http code is considered to be 200.
        healthy_http_codes = target_group_params.healthy_http_codes
        healthy_http_codes = [str(code) for code in healthy_http_codes] if healthy_http_codes else ['200']
        healthy_http_codes = ','.join(healthy_http_codes)

        health_check_path = target_group_params.health_check_path

        return CfnTargetGroup(
            self.__scope,
            target_group_params.prefix + 'TargetGroup',
            name=target_group_params.prefix + 'TargetGroup',
            matcher=CfnTargetGroup.MatcherProperty(http_code=healthy_http_codes),
            port=target_group_params.target_group_port,
            protocol=target_group_params.protocol,
            vpc_id=target_group_params.vpc.vpc_id,
            target_type=target_group_params.target_type,
            health_check_path=health_check_path if health_check_path else '/'
        )



from aws_cdk import aws_ec2
from aws_cdk.aws_ec2 import ISecurityGroup, IPeer


class SecurityGroupModifier:
    """
    Class which modifies a given security group.
    """
    def __init__(self, security_group: ISecurityGroup):
        """
        Constructor.

        :param security_group: Security group to modify.
        """
        self.__security_group = security_group

    def open_port(self, port: int, peer: IPeer, ingress: bool = True) -> None:
        """
        Modifies a given security group by opening a specified port.

        :param port: Port to open (allow traffic).
        :param peer: Peer (a CIDR or another security group).
        :param ingress: Specifies whether it is configured for ingress or egress traffic.

        :return: No return.
        """
        assert port is not None
        assert peer is not None
        assert ingress is not None

        sg = self.__security_group

        if ingress:
            sg.add_ingress_rule(
                peer=peer,
                connection=aws_ec2.Port(
                    protocol=aws_ec2.Protocol.TCP,
                    string_representation=f'Ingress {port} rule.',
                    from_port=port,
                    to_port=port
                )
            )
        else:
            sg.add_egress_rule(
                peer=peer,
                connection=aws_ec2.Port(
                    protocol=aws_ec2.Protocol.TCP,
                    string_representation=f'Egress {port} rule.',
                    from_port=port,
                    to_port=port
                )
            )



from aws_cdk.aws_elasticloadbalancingv2 import CfnListener
from aws_cdk.core import Stack
from aws_alb.listener_actions import ListenerActions
from aws_alb.params.listener_params import ListenerParams


class ListenerFactory:
    """
    Factory class which creates listeners for loadbalancer.
    """
    def __init__(self, scope: Stack) -> None:
        """
        Constructor.

        :param scope: CloudFormation stack instance.
        """
        self.__scope = scope

    def create(self, listener_params: ListenerParams) -> CfnListener:
        """
        Creates listener.

        :param listener_params: Configuration parameters on how to create the listener.

        :return: Listener instance.
        """
        assert listener_params.loadbalancer is not None, 'Loadbalancer must be specified.'
        assert listener_params.port is not None, 'Port must be specified.'

        protocol = 'HTTPS' if listener_params.certificate else 'HTTP'

        listener = CfnListener(
            scope=self.__scope,
            id=listener_params.prefix + f'{protocol.capitalize()}Listener{listener_params.port}',
            port=listener_params.port,
            protocol=protocol,
            load_balancer_arn=listener_params.loadbalancer.ref,
            default_actions=[listener_params.action or ListenerActions.fixed_404_action()]
        )

        if listener_params.certificate:
            listener.certificates = [
                CfnListener.CertificateProperty(certificate_arn=listener_params.certificate.ref)
            ]

        return listener



from aws_cdk.aws_elasticloadbalancingv2 import CfnListener, CfnTargetGroup


class ListenerActions:
    """
    Factory class to create various actions for loadbalancer listeners.
    """
    @staticmethod
    def redirect_action(port: int, protocol: str = 'HTTPS'):
        """
        Creates redirect action to same host with same path and query
        but to different port with different protocol.

        :param port: Port to redirect to.
        :param protocol: Protocol to use.

        :return: Redirect action.
        """
        return CfnListener.ActionProperty(
            type='redirect',
            redirect_config=CfnListener.RedirectConfigProperty(
                status_code='HTTP_301',
                host='#{host}',
                path='/#{path}',
                port=str(port),
                query='#{query}',
                protocol=protocol
            )
        )

    @staticmethod
    def target_group_action(target_group: CfnTargetGroup) -> CfnListener.ActionProperty:
        """
        Creates forward action to a target group.

        :param target_group: Target group to forward to.

        :return: Forward action.
        """
        return CfnListener.ActionProperty(
            type='forward',
            target_group_arn=target_group.ref
        )

    @staticmethod
    def fixed_404_action() -> CfnListener.ActionProperty:
        """
        Creates fixed 404 (not found) response.

        :return: Fixed 404 response action.
        """
        return CfnListener.ActionProperty(
            type='fixed-response',
            fixed_response_config=CfnListener.FixedResponseConfigProperty(
                status_code='404',
                message_body='Not found.'
            )
        )



from typing import Tuple
from aws_cdk import core
from aws_cdk.aws_certificatemanager import CfnCertificate
from aws_cdk.aws_elasticloadbalancingv2 import CfnListener, CfnTargetGroup, CfnLoadBalancer
from aws_alb.factories.listener_factory import ListenerFactory
from aws_alb.factories.target_group_factory import TargetGroupFactory
from aws_alb.listener_actions import ListenerActions
from aws_alb.params.listener_params import ListenerParams
from aws_alb.params.target_group_params import TargetGroupParams


class LoadBalancerListeners:
    """
    A class which manages listeners for a loadbalancer.
    """

    def __init__(
            self,
            scope: core.Stack,
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation stack in which the resources should be added.
        """
        self.__scope = scope
        self.__available_ports = list(range(20000, 30000))
        self.__listener_factory = ListenerFactory(scope)
        self.__target_group_factory = TargetGroupFactory(scope)

    def create_default_listeners(
            self,
            prefix: str,
            loadbalancer: CfnLoadBalancer,
            certificate: CfnCertificate
    ) -> Tuple[CfnListener, CfnListener]:
        """
        Creates default listeners for normal loadbalancer use.
        It creates two blue listeners that listen to 80 (http) and 443 (https) traffic.
        It creates two green listeners that listen to 8000 (http) and 44300 (https) traffic.
        Listener 80 redirects to 443 and listener 8000 redirects to 44300.

        :param prefix: A string prefix for resources.
        :param loadbalancer:
        :param certificate:

        :return: Tuple with two tuples. First tuple contains a blue pair and the second group contains a green pair.
        A pair contains two elements: a target group and a listener.
        """
        blue_https = self.create_listener(ListenerParams(
            prefix + 'BlueHttps',
            loadbalancer,
            certificate=certificate,
            port=443,
            action=ListenerActions.fixed_404_action()
        ))

        green_https = self.create_listener(ListenerParams(
            prefix + 'GreenHttps',
            loadbalancer,
            certificate=certificate,
            port=44300,
            action=ListenerActions.fixed_404_action()
        ))

        self.create_listener(ListenerParams(
            prefix + 'BlueHttp',
            loadbalancer,
            port=80,
            action=ListenerActions.redirect_action(blue_https.port)
        ))

        self.create_listener(ListenerParams(
            prefix + 'GreenHttp',
            loadbalancer,
            port=8000,
            action=ListenerActions.redirect_action(green_https.port)
        ))

        return blue_https, green_https

    def create_blue_green(
            self,
            blue_listener_params: ListenerParams,
            green_listener_params: ListenerParams,
            blue_target_group_params: TargetGroupParams,
            green_target_group_params: TargetGroupParams
    ) -> Tuple[Tuple[CfnTargetGroup, CfnListener], Tuple[CfnTargetGroup, CfnListener]]:
        """
        Creates listeners and target groups for out-of-the-box blue green deployments with AWS CodeDeploy.

        :param blue_listener_params: Listener parameters for the blue group.
        :param green_listener_params: Listener parameters for the green group.
        :param blue_target_group_params:  Target group parameters for the blue group.
        :param green_target_group_params: Target group parameters for the green group.

        :return: Tuple with two tuples. First tuple contains a blue pair and the second group contains a green pair.
        A pair contains two elements: a target group and a listener.
        """
        # Ensure ports are defined.
        blue_listener_params.port = blue_listener_params.port or self.__allocate_port()
        green_listener_params.port = green_listener_params.port or self.__allocate_port()

        # Ensure prefixes tell which is green and which is blue.
        blue_target_group_params.prefix = blue_target_group_params.prefix + 'Blue'
        blue_listener_params.prefix = blue_listener_params.prefix + 'Blue'
        green_target_group_params.prefix = green_target_group_params.prefix + 'Green'
        green_listener_params.prefix = green_listener_params.prefix + 'Green'

        # Create target groups and forward to them.
        blue_group = self.create_target_group(blue_target_group_params)
        blue_listener_params.action = ListenerActions.target_group_action(blue_group)
        green_group = self.create_target_group(green_target_group_params)
        green_listener_params.action = ListenerActions.target_group_action(green_group)

        # Create listeners.
        blue_listener = self.create_listener(blue_listener_params)
        green_listener = self.create_listener(green_listener_params)

        return (blue_group, blue_listener), (green_group, green_listener)

    def create_listener(self, listener_params: ListenerParams) -> CfnListener:
        """
        Creates a listener for a loadbalancer.

        :param listener_params: Configuration parameters for the new listener.

        :return: Listener instance.
        """
        sg = listener_params.loadbalancer_sg

        if sg:
            inbound = sg.get_peer(listener_params.inbound_traffic)
            outbound = sg.get_peer(listener_params.outbound_traffic)

            sg.open_port(listener_params.port, inbound, ingress=True)
            sg.open_port(listener_params.port, outbound, ingress=False)

        return self.__listener_factory.create(listener_params)

    def create_target_group(self, target_group_params: TargetGroupParams) -> CfnTargetGroup:
        """
        Creates a target group.

        :param target_group_params: Configuration parameters for the new target group.

        :return: Target group instance.
        """
        return self.__target_group_factory.create_target_group(target_group_params)

    def __allocate_port(self) -> int:
        """
        Allocates a free available port for a listener.

        :return: Available port.
        """
        return self.__available_ports.pop(0)



from typing import Optional, List
from aws_cdk.aws_ec2 import IVpc


class TargetGroupParams:
    """
    Parameters class for target group.
    """
    def __init__(
            self,
            prefix: str,
            vpc: IVpc,
            healthy_http_codes: Optional[List[int]] = None,
            health_check_path: Optional[str] = None,
            target_group_port: int = 80,
            protocol: str = 'HTTP',
            target_type: str = 'ip',
    ) -> None:
        """
        Constructor.

        :param prefix: String prefix for target group name.
        :param vpc: Virtual private cloud for the target group.
        :param healthy_http_codes: The deployed instance is constantly pinged to determine if it is available
        (healthy) or not. Specify a list of http codes that your service can return and should be treated as healthy.
        :param health_check_path: The deployed instance is constantly pinged to determine if it is available
        (healthy) or not. Specify a path for that ping.
        :param target_group_port: A port for a communication between loadbalancer and the target group.
        :param protocol: Protocol (http or https).
        :param target_type: Resource in the target group target type (ip or instance).
        """
        self.prefix = prefix
        self.vpc = vpc
        self.healthy_http_codes = healthy_http_codes
        self.health_check_path = health_check_path
        self.target_group_port = target_group_port
        self.protocol = protocol
        self.target_type = target_type



from typing import Optional

from aws_alb.loadbalancer_sg import LoadBalancerSecurityGroup
from aws_cdk.aws_certificatemanager import CfnCertificate
from aws_cdk.aws_elasticloadbalancingv2 import CfnListener, CfnLoadBalancer
from aws_alb.alb_traffic_enum import AlbTrafficEnum


class ListenerParams:
    """
    Parameters class for loadbalancer listeners.
    """
    def __init__(
            self,
            prefix: str,
            loadbalancer: CfnLoadBalancer,
            loadbalancer_sg: Optional[LoadBalancerSecurityGroup] = None,
            port: Optional[int] = None,
            inbound_traffic: AlbTrafficEnum = None,
            outbound_traffic: AlbTrafficEnum = None,
            certificate: Optional[CfnCertificate] = None,
            action: Optional[CfnListener.ActionProperty] = None,
    ) -> None:
        """
        Constructor.

        :param prefix: String prefix for listener name.
        :param loadbalancer: A loadbalancer for which the listener should be configured.
        :param loadbalancer_sg: A loadbalancer's security group.
        :param port: Port for listener to listen.
        :param inbound_traffic: Inbound traffic configuration (for security group).
        :param outbound_traffic: Outbound traffic configuration (for security group).
        :param certificate: Certificate to enable https traffic.
        :param action: Action to take when listener gets an incoming traffic.
        """
        self.prefix = prefix
        self.loadbalancer = loadbalancer
        self.loadbalancer_sg = loadbalancer_sg
        self.port = port
        self.inbound_traffic = inbound_traffic or AlbTrafficEnum.INTERNET
        self.outbound_traffic = outbound_traffic or AlbTrafficEnum.NONE
        self.certificate = certificate
        self.action = action
